from __future__ import annotations
import os

import asyncio
import platform
from typing import Any, Callable, ClassVar, Type, TypeVar
import warnings

from rich.control import Control
import rich.repr
from rich.screen import Screen
from rich.console import Console, RenderableType
from rich.measure import Measurement
from rich.traceback import Traceback

from . import events
from . import actions
from ._animator import Animator
from .binding import Bindings, NoBinding
from .geometry import Offset, Region
from . import log
from ._callback import invoke
from ._context import active_app
from ._event_broker import extract_handler_actions, NoHandler
from .driver import Driver
from .layouts.dock import DockLayout, Dock
from .message_pump import MessagePump
from ._profile import timer
from .view import View
from .views import DockView
from .widget import Widget, Reactive

PLATFORM = platform.system()
WINDOWS = PLATFORM == "Windows"

# asyncio will warn against resources not being cleared
warnings.simplefilter("always", ResourceWarning)


LayoutDefinition = "dict[str, Any]"

ViewType = TypeVar("ViewType", bound=View)

try:
    import uvloop
except ImportError:
    pass
else:
    uvloop.install()


class ActionError(Exception):
    pass


@rich.repr.auto
class App(MessagePump):
    """The base class for Textual Applications"""

    KEYS: ClassVar[dict[str, str]] = {}

    def __init__(
        self,
        screen: bool = True,
        driver_class: Type[Driver] | None = None,
        log: str = "",
        log_verbosity: int = 1,
        title: str = "Textual Application",
    ):
        """The Textual Application base class

        Args:
            console (Console, optional): A Rich Console. Defaults to None.
            screen (bool, optional): Enable full-screen application mode. Defaults to True.
            driver_class (Type[Driver], optional): Driver class, or None to use default. Defaults to None.
            title (str, optional): Title of the application. Defaults to "Textual Application".
        """
        self.console = Console()
        self.error_console = Console(stderr=True)
        self._screen = screen
        self.driver_class = driver_class or self.get_driver_class()
        self._title = title
        self._layout = DockLayout()
        self._view_stack: list[DockView] = []
        self.children: set[MessagePump] = set()

        self.focused: Widget | None = None
        self.mouse_over: Widget | None = None
        self.mouse_captured: Widget | None = None
        self._driver: Driver | None = None
        self._exit_renderables: list[RenderableType] = []

        self._docks: list[Dock] = []
        self._action_targets = {"app", "view"}
        self._animator = Animator(self)
        self.animate = self._animator.bind(self)
        self.mouse_position = Offset(0, 0)
        self.bindings = Bindings()
        self._title = title

        self.log_file = open(log, "wt") if log else None
        self.log_verbosity = log_verbosity

        self.bindings.bind("ctrl+c", "quit", show=False, allow_forward=False)
        self._refresh_required = False

        super().__init__()

    title: Reactive[str] = Reactive("Textual")
    sub_title: Reactive[str] = Reactive("")
    background: Reactive[str] = Reactive("black")

    def get_driver_class(self) -> Type[Driver]:
        """Get a driver class for this platform.

        Called by the constructor.

        Returns:
            Driver: A Driver class which manages input and display.
        """
        if WINDOWS:
            from .drivers.windows_driver import WindowsDriver

            driver_class = WindowsDriver
        else:
            from .drivers.linux_driver import LinuxDriver

            driver_class = LinuxDriver
        return driver_class

    def __rich_repr__(self) -> rich.repr.Result:
        yield "title", self.title

    def __rich__(self) -> RenderableType:
        return self.view

    @property
    def animator(self) -> Animator:
        return self._animator

    @property
    def view(self) -> DockView:
        return self._view_stack[-1]

    def log(self, *args: Any, verbosity: int = 1, **kwargs) -> None:
        """Write to logs.

        Args:
            *args (Any): Positional arguments are converted to string and written to logs.
            verbosity (int, optional): Verbosity level 0-3. Defaults to 1.
        """
        try:
            if self.log_file and verbosity <= self.log_verbosity:
                output = f" ".join(str(arg) for arg in args)
                if kwargs:
                    key_values = " ".join(
                        f"{key}={value}" for key, value in kwargs.items()
                    )
                    output = " ".join((output, key_values))
                self.log_file.write(output + "\n")
                self.log_file.flush()
        except Exception:
            pass

    async def bind(
        self,
        keys: str,
        action: str,
        description: str = "",
        show: bool = True,
        key_display: str | None = None,
    ) -> None:
        """Bind a key to an action.

        Args:
            keys (str): A comma separated list of keys, i.e.
            action (str): Action to bind to.
            description (str, optional): Short description of action. Defaults to "".
            show (bool, optional): Show key in UI. Defaults to True.
            key_display (str, optional): Replacement text for key, or None to use default. Defaults to None.
        """
        self.bindings.bind(
            keys, action, description, show=show, key_display=key_display
        )

    @classmethod
    def run(
        cls,
        console: Console = None,
        screen: bool = True,
        driver: Type[Driver] = None,
        **kwargs,
    ):
        """Run the app.

        Args:
            console (Console, optional): Console object. Defaults to None.
            screen (bool, optional): Enable application mode. Defaults to True.
            driver (Type[Driver], optional): Driver class or None for default. Defaults to None.
        """

        async def run_app() -> None:
            app = cls(screen=screen, driver_class=driver, **kwargs)
            await app.process_messages()

        asyncio.run(run_app())

    async def push_view(self, view: ViewType) -> ViewType:
        self.register(view, self)
        self._view_stack.append(view)
        return view

    async def set_focus(self, widget: Widget | None) -> None:
        """Focus (or unfocus) a widget. A focused widget will receive key events first.

        Args:
            widget (Widget): [description]
        """
        log("set_focus", widget)
        if widget == self.focused:
            # Widget is already focused
            return

        if widget is None:
            if self.focused is not None:
                focused = self.focused
                self.focused = None
                await focused.post_message(events.Blur(self))
        elif widget.can_focus:
            if self.focused is not None:
                await self.focused.post_message(events.Blur(self))
            if widget is not None and self.focused != widget:
                self.focused = widget
                await widget.post_message(events.Focus(self))

    async def set_mouse_over(self, widget: Widget | None) -> None:
        if widget is None:
            if self.mouse_over is not None:
                try:
                    await self.mouse_over.post_message(events.Leave(self))
                finally:
                    self.mouse_over = None
        else:
            if self.mouse_over != widget:
                try:
                    if self.mouse_over is not None:
                        await self.mouse_over.forward_event(events.Leave(self))
                    if widget is not None:
                        await widget.forward_event(events.Enter(self))
                finally:
                    self.mouse_over = widget

    async def capture_mouse(self, widget: Widget | None) -> None:
        """Send all mouse events to the given widget, disable mouse capture.

        Args:
            widget (Widget | None): If a widget, capture mouse event, or None to end mouse capture.
        """
        if widget == self.mouse_captured:
            return
        if self.mouse_captured is not None:
            await self.mouse_captured.post_message(
                events.MouseRelease(self, self.mouse_position)
            )
        self.mouse_captured = widget
        if widget is not None:
            await widget.post_message(events.MouseCapture(self, self.mouse_position))

    def panic(self, *renderables: RenderableType) -> None:
        """Exits the app with a traceback.

        Args:
            traceback (Traceback, optional): Rich Traceback object or None to generate one
                for the most recent exception. Defaults to None.
        """

        if not renderables:
            renderables = (
                Traceback(show_locals=True, width=None, locals_max_length=5),
            )
        self._exit_renderables.extend(renderables)
        self.close_messages_no_wait()

    async def process_messages(self) -> None:
        active_app.set(self)

        log("---")
        log(f"driver={self.driver_class}")

        load_event = events.Load(sender=self)
        await self.dispatch_message(load_event)
        await self.post_message(events.Mount(self))
        await self.push_view(DockView())

        # Wait for the load event to be processed, so we don't go in to application mode beforehand
        await load_event.wait()

        driver = self._driver = self.driver_class(self.console, self)
        try:
            driver.start_application_mode()
        except Exception:
            self.console.print_exception()
        else:
            try:
                self.console = Console()
                self.title = self._title
                self.refresh()
                await self.animator.start()
                await super().process_messages()
                log("PROCESS END")
                await self.animator.stop()
                await self.close_all()

            except Exception:
                self.panic()
            finally:
                driver.stop_application_mode()
                if self._exit_renderables:
                    for renderable in self._exit_renderables:
                        self.error_console.print(renderable)
                if self.log_file is not None:
                    self.log_file.close()

    def register(self, child: MessagePump, parent: MessagePump) -> bool:
        if child not in self.children:
            self.children.add(child)
            child.set_parent(parent)
            child.start_messages()
            child.post_message_no_wait(events.Mount(sender=parent))
            return True
        return False

    async def close_all(self) -> None:
        while self.children:
            child = self.children.pop()
            await child.close_messages()

    async def remove(self, child: MessagePump) -> None:
        self.children.remove(child)

    async def shutdown(self):
        driver = self._driver
        assert driver is not None
        driver.disable_input()
        await self.close_messages()

    def refresh(self, repaint: bool = True, layout: bool = False) -> None:
        sync_available = (
            os.environ.get("TERM_PROGRAM", "") != "Apple_Terminal" and not WINDOWS
        )
        if not self._closed:
            console = self.console
            try:
                if sync_available:
                    console.file.write("\x1bP=1s\x1b\\")
                console.print(Screen(Control.home(), self.view, Control.home()))
                if sync_available:
                    console.file.write("\x1bP=2s\x1b\\")
                console.file.flush()
            except Exception:
                self.panic()

    def display(self, renderable: RenderableType) -> None:
        sync_available = os.environ.get("TERM_PROGRAM", "") != "Apple_Terminal"
        if not self._closed:
            console = self.console
            try:
                console.print(renderable)
            except Exception:
                self.panic()

    def measure(self, renderable: RenderableType, max_width=100_000) -> int:
        """Get the optimal width for a widget or renderable.

        Args:
            renderable (RenderableType): A renderable (including Widget)
            max_width ([type], optional): Maximum width. Defaults to 100_000.

        Returns:
            int: Number of cells required to render.
        """
        measurement = Measurement.get(
            self.console, self.console.options.update(max_width=max_width), renderable
        )
        return measurement.maximum

    def get_widget_at(self, x: int, y: int) -> tuple[Widget, Region]:
        """Get the widget under the given coordinates.

        Args:
            x (int): X Coord.
            y (int): Y Coord.

        Returns:
            tuple[Widget, Region]: The widget and the widget's screen region.
        """
        return self.view.get_widget_at(x, y)

    async def press(self, key: str) -> bool:
        """Handle a key press.

        Args:
            key (str): A key

        Returns:
            bool: True if the key was handled by a binding, otherwise False
        """
        try:
            binding = self.bindings.get_key(key)
        except NoBinding:
            return False
        else:
            await self.action(binding.action)
        return True

    async def on_event(self, event: events.Event) -> None:
        # Handle input events that haven't been forwarded
        # If the event has been forwarded it may have bubbled up back to the App
        if isinstance(event, events.InputEvent) and not event.is_forwarded:
            if isinstance(event, events.MouseEvent):
                # Record current mouse position on App
                self.mouse_position = Offset(event.x, event.y)
            if isinstance(event, events.Key) and self.focused is not None:
                # Key events are sent direct to focused widget
                if self.bindings.allow_forward(event.key):
                    await self.focused.forward_event(event)
                else:
                    # Key has allow_forward=False which disallows forward to focused widget
                    await super().on_event(event)
            else:
                # Forward the event to the view
                await self.view.forward_event(event)
        else:
            await super().on_event(event)

    async def action(
        self,
        action: str,
        default_namespace: object | None = None,
        modifiers: set[str] | None = None,
    ) -> None:
        """Perform an action.

        Args:
            action (str): Action encoded in a string.
        """
        target, params = actions.parse(action)
        if "." in target:
            destination, action_name = target.split(".", 1)
            if destination not in self._action_targets:
                raise ActionError("Action namespace {destination} is not known")
            action_target = getattr(self, destination)
        else:
            action_target = default_namespace or self
            action_name = target

        log("ACTION", action_target, action_name)
        await self.dispatch_action(action_target, action_name, params)

    async def dispatch_action(
        self, namespace: object, action_name: str, params: Any
    ) -> None:
        _rich_traceback_guard = True
        method_name = f"action_{action_name}"
        method = getattr(namespace, method_name, None)
        if callable(method):
            await invoke(method, *params)

    async def broker_event(
        self, event_name: str, event: events.Event, default_namespace: object | None
    ) -> bool:
        event.stop()
        try:
            style = getattr(event, "style")
        except AttributeError:
            return False
        try:
            modifiers, action = extract_handler_actions(event_name, style.meta)
        except NoHandler:
            return False
        if isinstance(action, str):
            await self.action(
                action, default_namespace=default_namespace, modifiers=modifiers
            )
        elif isinstance(action, Callable):
            await action()
        else:
            return False
        return True

    async def on_key(self, event: events.Key) -> None:
        await self.press(event.key)

    async def on_shutdown_request(self, event: events.ShutdownRequest) -> None:
        log("shutdown request")
        await self.close_messages()

    async def on_resize(self, event: events.Resize) -> None:
        await self.view.post_message(event)

    async def action_press(self, key: str) -> None:
        await self.press(key)

    async def action_quit(self) -> None:
        await self.shutdown()

    async def action_bang(self) -> None:
        1 / 0

    async def action_bell(self) -> None:
        self.console.bell()


if __name__ == "__main__":
    import asyncio
    from logging import FileHandler

    from rich.panel import Panel

    from .widgets import Header
    from .widgets import Footer

    from .widgets import Placeholder
    from .scrollbar import ScrollBar

    from rich.markdown import Markdown

    # from .widgets.scroll_view import ScrollView

    import os

    class MyApp(App):
        """Just a test app."""

        async def on_load(self, event: events.Load) -> None:
            await self.bind("ctrl+c", "quit", show=False)
            await self.bind("q", "quit", "Quit")
            await self.bind("x", "bang", "Test error handling")
            await self.bind("b", "toggle_sidebar", "Toggle sidebar")

        show_bar: Reactive[bool] = Reactive(False)

        async def watch_show_bar(self, show_bar: bool) -> None:
            self.animator.animate(self.bar, "layout_offset_x", 0 if show_bar else -40)

        async def action_toggle_sidebar(self) -> None:
            self.show_bar = not self.show_bar

        async def on_mount(self, event: events.Mount) -> None:

            view = await self.push_view(DockView())

            header = Header()
            footer = Footer()
            self.bar = Placeholder(name="left")

            await view.dock(header, edge="top")
            await view.dock(footer, edge="bottom")
            await view.dock(self.bar, edge="left", size=40, z=1)
            self.bar.layout_offset_x = -40

            sub_view = DockView()
            await sub_view.dock(Placeholder(), Placeholder(), edge="top")
            await view.dock(sub_view, edge="left")

    MyApp.run(log="textual.log")



from __future__ import annotations

from rich import box
from rich.align import Align
from rich.console import RenderableType
from rich.panel import Panel
from rich.pretty import Pretty
import rich.repr

from logging import getLogger

from .. import events
from ..geometry import Offset
from ..widget import Reactive, Widget

log = getLogger("rich")


@rich.repr.auto(angular=False)
class Placeholder(Widget, can_focus=True):

    has_focus: Reactive[bool] = Reactive(False)
    mouse_over: Reactive[bool] = Reactive(False)
    style: Reactive[str] = Reactive("")
    height: Reactive[int | None] = Reactive(None)

    def __init__(self, *, name: str | None = None, height: int | None = None) -> None:
        super().__init__(name=name)
        self.height = height

    def __rich_repr__(self) -> rich.repr.Result:
        yield "name", self.name
        yield "has_focus", self.has_focus, False
        yield "mouse_over", self.mouse_over, False

    def render(self) -> RenderableType:
        return Panel(
            Align.center(
                Pretty(self, no_wrap=True, overflow="ellipsis"), vertical="middle"
            ),
            title=self.__class__.__name__,
            border_style="green" if self.mouse_over else "blue",
            box=box.HEAVY if self.has_focus else box.ROUNDED,
            style=self.style,
            height=self.height,
        )

    async def on_focus(self, event: events.Focus) -> None:
        self.has_focus = True

    async def on_blur(self, event: events.Blur) -> None:
        self.has_focus = False

    async def on_enter(self, event: events.Enter) -> None:
        self.mouse_over = True

    async def on_leave(self, event: events.Leave) -> None:
        self.mouse_over = False



from ._footer import Footer
from ._header import Header
from ._button import Button, ButtonPressed
from ._placeholder import Placeholder
from ._scroll_view import ScrollView
from ._static import Static
from ._tree_control import TreeControl, TreeClick, TreeNode, NodeID
from ._directory_tree import DirectoryTree, FileClick

__all__ = [
    "Button",
    "ButtonPressed",
    "DirectoryTree",
    "FileClick",
    "Footer",
    "Header",
    "Placeholder",
    "ScrollView",
    "Static",
    "TreeClick",
    "TreeControl",
    "TreeNode",
    "NodeID",
]



from __future__ import annotations

import sys
from collections import defaultdict
from dataclasses import dataclass
from typing import Iterable, TYPE_CHECKING, Sequence

from rich.console import Console

from .._layout_resolve import layout_resolve
from ..geometry import Offset, Region, Size
from ..layout import Layout, WidgetPlacement
from ..layout_map import LayoutMap

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


if TYPE_CHECKING:
    from ..widget import Widget


DockEdge = Literal["top", "right", "bottom", "left"]


@dataclass
class DockOptions:
    size: int | None = None
    fraction: int = 1
    min_size: int = 1


@dataclass
class Dock:
    edge: DockEdge
    widgets: Sequence[Widget]
    z: int = 0


class DockLayout(Layout):
    def __init__(self, docks: list[Dock] = None) -> None:
        self.docks: list[Dock] = docks or []
        super().__init__()

    def get_widgets(self) -> Iterable[Widget]:
        for dock in self.docks:
            yield from dock.widgets

    def arrange(self, size: Size, scroll: Offset) -> Iterable[WidgetPlacement]:

        map: LayoutMap = LayoutMap(size)
        width, height = size
        layout_region = Region(0, 0, width, height)
        layers: dict[int, Region] = defaultdict(lambda: layout_region)

        for index, dock in enumerate(self.docks):
            dock_options = [
                DockOptions(
                    widget.layout_size, widget.layout_fraction, widget.layout_min_size
                )
                for widget in dock.widgets
            ]
            region = layers[dock.z]
            if not region:
                # No space left
                continue

            order = (dock.z, index)
            x, y, width, height = region

            if dock.edge == "top":
                sizes = layout_resolve(height, dock_options)
                render_y = y
                remaining = region.height
                total = 0
                for widget, layout_size in zip(dock.widgets, sizes):
                    if not widget.visible:
                        continue
                    layout_size = min(remaining, layout_size)
                    if not layout_size:
                        break
                    total += layout_size
                    yield WidgetPlacement(
                        Region(x, render_y, width, layout_size), widget, order
                    )
                    render_y += layout_size
                    remaining = max(0, remaining - layout_size)
                region = Region(x, y + total, width, height - total)

            elif dock.edge == "bottom":
                sizes = layout_resolve(height, dock_options)
                render_y = y + height
                remaining = region.height
                total = 0
                for widget, layout_size in zip(dock.widgets, sizes):
                    if not widget.visible:
                        continue
                    layout_size = min(remaining, layout_size)
                    if not layout_size:
                        break
                    total += layout_size
                    yield WidgetPlacement(
                        Region(x, render_y - layout_size, width, layout_size),
                        widget,
                        order,
                    )
                    render_y -= layout_size
                    remaining = max(0, remaining - layout_size)
                region = Region(x, y, width, height - total)

            elif dock.edge == "left":
                sizes = layout_resolve(width, dock_options)
                render_x = x
                remaining = region.width
                total = 0
                for widget, layout_size in zip(dock.widgets, sizes):
                    if not widget.visible:
                        continue
                    layout_size = min(remaining, layout_size)
                    if not layout_size:
                        break
                    total += layout_size
                    yield WidgetPlacement(
                        Region(render_x, y, layout_size, height),
                        widget,
                        order,
                    )
                    render_x += layout_size
                    remaining = max(0, remaining - layout_size)
                region = Region(x + total, y, width - total, height)

            elif dock.edge == "right":
                sizes = layout_resolve(width, dock_options)
                render_x = x + width
                remaining = region.width
                total = 0
                for widget, layout_size in zip(dock.widgets, sizes):
                    if not widget.visible:
                        continue
                    layout_size = min(remaining, layout_size)
                    if not layout_size:
                        break
                    total += layout_size
                    yield WidgetPlacement(
                        Region(render_x - layout_size, y, layout_size, height),
                        widget,
                        order,
                    )
                    render_x -= layout_size
                    remaining = max(0, remaining - layout_size)
                region = Region(x, y, width - total, height)

            layers[dock.z] = region

        return map



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['textual',
 'textual.drivers',
 'textual.layouts',
 'textual.views',
 'textual.widgets']

package_data = \
{'': ['*']}

install_requires = \
['rich>=12.0.0,<13.0.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.10.0,<4.0.0']}

setup_kwargs = {
    'name': 'textual',
    'version': '0.1.17',
    'description': 'Text User Interface using Rich',
    'long_description': '# Textual\n\n![screenshot](./imgs/textual.png)\n\nTextual is a TUI (Text User Interface) framework for Python inspired by modern web development. Currently a **Work in Progress**.\n\n\n> ⚠ **NOTE:** We ([Textualize.io](https://www.textualize.io)) are hard at work on the **css** branch. We will maintain the 0.1.0 branch for the near future but may not be able to accept API changes. If you would like to contribute code via a PR, please raise a discussion first, to avoid disapointment.\n\n\nFollow [@willmcgugan](https://twitter.com/willmcgugan) for progress updates, or post in Discussions if you have any requests / suggestions.\n\n[![Join the chat at https://gitter.im/textual-ui/community](https://badges.gitter.im/textual-ui/community.svg)](https://gitter.im/textual-ui/community?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n\n## Compatibility\n\nTextual currently runs on **MacOS / Linux / Window**.\n\n## How it works\n\nTextual uses [Rich](https://github.com/willmcgugan/rich) to render rich text, so anything that Rich can render may be used in Textual.\n\nEvent handling in Textual is asynchronous (using `async` and `await` keywords). Widgets (UI components) can independently update and communicate with each other via message passing.\n\nTextual has more in common with modern web development than it does with [curses](<https://en.wikipedia.org/wiki/Curses_(programming_library)>); layout is done with CSS grid and (soon) the theme may be customized with CSS. Other techniques are borrowed from JS frameworks such as Vue and React.\n\n## Installation\n\nYou can install Textual via pip (`pip install textual`), or by checking out the repo and installing with [poetry](https://python-poetry.org/docs/).\n\n```\npoetry install\n```\n\nOnce installed you can run the following command for a quick test, or see examples (below):\n\n```\npython -m textual.app\n```\n\nTextual requires Python 3.7 or above.\n\n## Examples\n\nUntil I\'ve written the documentation, the [examples](https://github.com/willmcgugan/textual/tree/main/examples/) may be the best way to learn Textual.\n\nYou can see some of these examples in action in the [Developer Video Log](#developer-video-log).\n\n- [animation.py](https://github.com/willmcgugan/textual/tree/main/examples/animation.py) Demonstration of 60fps animation easing function\n- [calculator.py](https://github.com/willmcgugan/textual/tree/main/examples/calculator.py) A "clone" of the MacOS calculator using Grid layout\n- [code_viewer.py](https://github.com/willmcgugan/textual/tree/main/examples/code_viewer.py) A demonstration of a tree view which loads syntax highlighted code\n- [grid.py](https://github.com/willmcgugan/textual/tree/main/examples/grid.py) A simple demonstration of adding widgets in a Grid layout\n- [grid_auto.py](https://github.com/willmcgugan/textual/tree/main/examples/grid_auto.py) A demonstration of automatic Grid layout\n- [simple.py](https://github.com/willmcgugan/textual/tree/main/examples/simple.py) A very simple Textual app with scrolling Markdown view\n\n## Building Textual applications\n\n_This guide is a work in progress_\n\nLet\'s look at the simplest Textual app which does _something_:\n\n```python\nfrom textual.app import App\n\n\nclass Beeper(App):\n    def on_key(self):\n        self.console.bell()\n\n\nBeeper.run()\n```\n\nHere we can see a textual app with a single `on_key` method which will handle key events. Pressing any key will result in playing the terminal bell (generally an irritating beep). Hit Ctrl+C to exit.\n\nEvent handlers in Textual are defined by convention, not by inheritance (there\'s no base class with all the handlers defined). Each event has a `name` attribute which for the key event is simply `"key"`. Textual will call the method named `on_<event.name>` if it exists.\n\nLet\'s look at a _slightly_ more interesting example:\n\n```python\nfrom textual.app import App\n\n\nclass ColorChanger(App):\n    def on_key(self, event):\n        if event.key.isdigit():\n            self.background = f"on color({event.key})"\n\n\nColorChanger.run(log="textual.log")\n```\n\nYou\'ll notice that the `on_key` method above contains an additional `event` parameter which wasn\'t present on the beeper example. If the `event` argument is present, Textual will call the handler with an event object. Every event has an associated handler object, in this case it is a KeyEvent which contains additional information regarding which key was pressed.\n\nThe key event handler above will set the background attribute if you press the keys 0-9, which turns the terminal to the corresponding [ansi color](https://rich.readthedocs.io/en/latest/appendix/colors.html).\n\nNote that we didn\'t need to explicitly refresh the screen or draw anything. Setting the `background` attribute to a [Rich style](https://rich.readthedocs.io/en/latest/style.html) is enough for Textual to update the visuals. This is an example of _reactivity_ in Textual. To make changes to the terminal interface you modify the _state_ and let Textual update the UI.\n\n## Widgets\n\nTo make more interesting apps you will need to make use of _widgets_, which are independent user interface elements. Textual comes with a (growing) library of widgets, but you can develop your own.\n\nLet\'s look at an app which contains widgets. We will be using the built-in `Placeholder` widget which you can use to design application layouts before you implement the real content.\n\n```python\nfrom textual.app import App\nfrom textual.widgets import Placeholder\n\n\nclass SimpleApp(App):\n\n    async def on_mount(self) -> None:\n        await self.view.dock(Placeholder(), edge="left", size=40)\n        await self.view.dock(Placeholder(), Placeholder(), edge="top")\n\n\nSimpleApp.run(log="textual.log")\n```\n\nThis app contains a single event handler `on_mount`. The mount event is sent when the app or widget is ready to start processing events, and is typically used for initialization. You may have noticed that `on_mount` is an `async` function. Since Textual is an asynchronous framework we will need this if we need to call most other methods.\n\nThe `on_mount` method makes two calls to `self.view.dock` which adds widgets to the terminal.\n\nHere\'s the first line in the mount handler:\n\n```python\nawait self.view.dock(Placeholder(), edge="left", size=40)\n```\n\nNote this method is asynchronous like almost all API methods in Textual. We are awaiting `self.view.dock` which takes a newly constructed Placeholder widget, and docks it on to the `"left"` edge of the terminal with a size of 40 characters. In a real app you might use this to display a side-bar.\n\nThe following line is similar:\n\n```python\nawait self.view.dock(Placeholder(), Placeholder(), edge="top")\n```\n\nYou will notice that this time we are docking _two_ Placeholder objects onto the `"top"` edge. We haven\'t set an explicit size this time so Textual will divide the remaining size amongst the two new widgets.\n\nThe last line calls the `run` class method in the usual way, but with an argument we haven\'t seen before: `log="textual.log"` tells Textual to write log information to the given file. You can tail textual.log to see events being processed and other debug information.\n\nIf you run the above example, you will see something like the following:\n\n![widgets](./imgs/widgets.png)\n\nIf you move the mouse over the terminal you will notice that widgets receive mouse events. You can click any of the placeholders to give it input focus.\n\nThe dock layout feature is very flexible, but for more sophisticated layouts we can use the grid API. See the [calculator.py](https://github.com/willmcgugan/textual/blob/main/examples/calculator.py) example which makes use of Grid.\n\n### Creating Widgets\n\nYou can create your own widgets by subclassing the `textual.widget.Widget` class and implementing a `render()` method which should return anything that can be rendered with [Rich](https://rich.readthedocs.io/en/latest/introduction.html), including a plain string which will be interpreted as [console markup](https://rich.readthedocs.io/en/latest/markup.html).\n\nLet\'s look at an example with a custom widget:\n\n```python\nfrom rich.panel import Panel\n\nfrom textual.app import App\nfrom textual.reactive import Reactive\nfrom textual.widget import Widget\n\n\nclass Hover(Widget):\n\n    mouse_over = Reactive(False)\n\n    def render(self) -> Panel:\n        return Panel("Hello [b]World[/b]", style=("on red" if self.mouse_over else ""))\n\n    def on_enter(self) -> None:\n        self.mouse_over = True\n\n    def on_leave(self) -> None:\n        self.mouse_over = False\n\n\nclass HoverApp(App):\n    """Demonstrates custom widgets"""\n\n    async def on_mount(self) -> None:\n        hovers = (Hover() for _ in range(10))\n        await self.view.dock(*hovers, edge="top")\n\n\nHoverApp.run(log="textual.log")\n```\n\nThe `Hover` class is a custom widget which displays a panel containing the classic text "Hello World". The first line in the Hover class may seem a little mysterious at this point:\n\n```python\nmouse_over = Reactive(False)\n```\n\nThis adds a `mouse_over` attribute to your class which is a bool with a default of `False`. Adding attributes like this makes them _reactive_: any changes will result in the widget updating.\n\nThe following `render()` method is where you define how the widget should be displayed. In the Hover widget we return a [Panel](https://rich.readthedocs.io/en/latest/panel.html) containing rich text with a background that changes depending on the value of `mouse_over`. The goal here is to add a mouse hover effect to the widget, which we can achieve by handling two events: `Enter` and `Leave`. These events are sent when the mouse enters or leaves the widget.\n\nHere are the two event handlers again:\n\n```python\n    def on_enter(self) -> None:\n        self.mouse_over = True\n\n    def on_leave(self) -> None:\n        self.mouse_over = False\n```\n\nBoth event handlers set the `mouse_over` attribute which will result in the widget\'s `render()` method being called.\n\nThe `HoverApp` has a `on_mount` handler which creates 10 Hover widgets and docks them on the top edge to create a vertical stack:\n\n```python\n    async def on_mount(self) -> None:\n        hovers = (Hover() for _ in range(10))\n        await self.view.dock(*hovers, edge="top")\n```\n\nIf you run this script you will see something like the following:\n\n![widgets](./imgs/custom.gif)\n\nIf you move your mouse over the terminal you should see that the widget under the mouse cursor changes to a red background.\n\n### Actions and key bindings\n\nActions in Textual are white-listed functions that may be bound to keys. Let\'s look at a trivial example of binding a key to an action. Here is an app which exits when we hit the Q key:\n\n```python\nfrom textual.app import App\n\n\nclass Quitter(App):\n    async def on_load(self, event):\n        await self.bind("q", "quit")\n\n\nQuitter.run()\n```\n\nIf you run this you will get a blank terminal which will return to the prompt when you press Q.\n\nBinding is done in the Load event handler. The `bind` method takes the key (in this case "q") and binds it to an action ("quit"). The quit action is built in to Textual and simply exits the app.\n\nTo define your own actions, add a method that begins with `action_`, which may take parameters. Let\'s create a simple action that changes the color of the terminal and binds keys to it:\n\n```python\nfrom textual.app import App\n\n\nclass Colorizer(App):\n\n    async def on_load(self, event):\n        await self.bind("r", "color(\'red\')")\n        await self.bind("g", "color(\'green\')")\n        await self.bind("b", "color(\'blue\')")\n\n    async def action_color(self, color:str) -> None:\n        self.background = f"on {color}"\n\n\nColorizer.run()\n```\n\nIf you run this app you can hit the keys R, G, or B to change the color of the background.\n\nIn the `on_load` method we have bound the keys R, G, and B to the `color` action with a single parameter. When you press any of these three keys Textual will call the method `action_color` with the appropriate parameter.\n\nYou could be forgiven for thinking that `"color(\'red\')"` is Python code which Textual evaluates. This is not the case. The action strings are parsed and may not include expressions or arbitrary code. The reason that strings are used over a callable is that (in a future update) key bindings may be loaded from a configuration file.\n\n### More on Events\n\n_TODO_\n\n### Watchers\n\n_TODO_\n\n### Animation\n\n_TODO_\n\n### Timers and Intervals\n\nTextual has a `set_timer` and a `set_interval` method which work much like their Javascript counterparts. The `set_timer` method will invoke a callable after a given period of time, and `set_interval` will invoke a callable repeatedly. Unlike Javascript these methods expect the time to be in seconds (_not_ milliseconds).\n\nLet\'s create a simple terminal based clock with the `set_interval` method:\n\n```python\nfrom datetime import datetime\n\nfrom rich.align import Align\n\nfrom textual.app import App\nfrom textual.widget import Widget\n\n\nclass Clock(Widget):\n    def on_mount(self):\n        self.set_interval(1, self.refresh)\n\n    def render(self):\n        time = datetime.now().strftime("%c")\n        return Align.center(time, vertical="middle")\n\n\nclass ClockApp(App):\n    async def on_mount(self):\n        await self.view.dock(Clock())\n\n\nClockApp.run()\n\n```\n\nIf you run this app you will see the current time in the center of the terminal until you hit Ctrl+C.\n\nThe Clock widget displays the time using [rich.align.Align](https://rich.readthedocs.io/en/latest/reference/align.html) to position it in the center. In the clock\'s Mount handler there is the following call to `set_interval`:\n\n```python\nself.set_interval(1, self.refresh)\n```\n\nThis tells Textual to call a function (in this case `self.refresh` which updates the widget) once a second. When a widget is refreshed it calls `Clock.render` again to display the latest time.\n\n## Developer Video Log\n\nSince Textual is a visual medium, I\'ll be documenting new features and milestones here.\n\n### Update 1 - Basic scrolling\n\n[![Textual update 1](https://yt-embed.herokuapp.com/embed?v=zNW7U36GHlU&img=0)](http://www.youtube.com/watch?v=zNW7U36GHlU)\n\n### Update 2 - Keyboard toggle\n\n[![Textual update 2](https://yt-embed.herokuapp.com/embed?v=bTYeFOVNXDI&img=0)](http://www.youtube.com/watch?v=bTYeFOVNXDI)\n\n### Update 3 - New scrollbars and smooth scrolling\n\n[![Textual update 3](https://yt-embed.herokuapp.com/embed?v=4LVl3ClrXIs&img=0)](http://www.youtube.com/watch?v=4LVl3ClrXIs)\n\n### Update 4 - Animation system with easing function\n\nNow with a system to animate changes to values, going from the initial to the final value in small increments over time . Here applied to the scroll position. The animation system supports CSS like _easing functions_. You may be able to tell from the video that the page up / down keys cause the window to first speed up and then slow down.\n\n[![Textual update 4](https://yt-embed.herokuapp.com/embed?v=k2VwOp1YbSk&img=0)](http://www.youtube.com/watch?v=k2VwOp1YbSk)\n\n### Update 5 - New Layout system\n\nA new update system allows for overlapping layers. Animation is now synchronized with the display which makes it very smooth!\n\n[![Textual update 5](https://yt-embed.herokuapp.com/embed?v=XxRnfx2WYRw&img=0)](http://www.youtube.com/watch?v=XxRnfx2WYRw)\n\n### Update 6 - New Layout API\n\nNew version (0.1.4) with API updates and the new layout system.\n\n[![Textual update 6](https://yt-embed.herokuapp.com/embed?v=jddccDuVd3E&img=0)](http://www.youtube.com/watch?v=jddccDuVd3E)\n\n### Update 7 - New Grid Layout\n\n**11 July 2021**\n\nAdded a new layout system modelled on CSS grid. The example demonstrates how once created a grid will adapt to the available space.\n\n[![Textual update 7](https://yt-embed.herokuapp.com/embed?v=Zh9CEvu73jc&img=0)](http://www.youtube.com/watch?v=Zh9CEvu73jc)\n\n## Update 8 - Tree control and scroll views\n\n**6 Aug 2021**\n\nAdded a tree control and refactored the renderer to allow for widgets within a scrollable view\n\n[![Textual update 8](https://yt-embed.herokuapp.com/embed?v=J-dzzD6NQJ4&img=0)](http://www.youtube.com/watch?v=J-dzzD6NQJ4)\n',
    'author': 'Will McGugan',
    'author_email': 'willmcgugan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/willmcgugan/textual',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations


from typing import Iterable

from rich.cells import cell_len
from rich.console import Console, ConsoleOptions, RenderableType, RenderResult
from rich.control import Control
from rich.segment import Segment
from rich.style import Style

from ._loop import loop_last


class LineCache:
    def __init__(self, lines: list[list[Segment]]) -> None:
        self.lines = lines
        self._dirty = [True] * len(self.lines)

    @classmethod
    def from_renderable(
        cls,
        console: Console,
        renderable: RenderableType,
        width: int,
        height: int,
    ) -> "LineCache":
        options = console.options.update_dimensions(width, height)
        lines = console.render_lines(renderable, options)
        return cls(lines)

    @property
    def dirty(self) -> bool:
        return any(self._dirty)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        new_line = Segment.line()
        for line in self.lines:
            yield from line
            yield new_line

    def render(self, x: int, y: int, width: int, height: int) -> Iterable[Segment]:
        move_to = Control.move_to
        lines = self.lines[:height]
        new_line = Segment.line()
        for last, (offset_y, (line, dirty)) in loop_last(
            enumerate(zip(lines, self._dirty), y)
        ):
            if dirty:
                yield move_to(x, offset_y).segment
                yield from Segment.adjust_line_length(line, width)
                if not last:
                    yield new_line
        self._dirty[:] = [False] * len(self.lines)

    def get_style_at(self, x: int, y: int) -> Style:
        try:
            line = self.lines[y]
        except IndexError:
            return Style.null()
        end = 0
        for segment in line:
            end += cell_len(segment.text)
            if x < end:
                return segment.style or Style.null()
        return Style.null()



from ..view import View
from ..layouts.grid import GridLayout


class GridView(View, layout=GridLayout):
    @property
    def grid(self) -> GridLayout:
        assert isinstance(self.layout, GridLayout)
        return self.layout



from __future__ import annotations

from rich.console import RenderableType

from .. import events
from ..geometry import Size, SpacingDimensions
from ..layouts.vertical import VerticalLayout
from ..view import View
from ..message import Message
from .. import messages
from ..widget import Widget
from ..widgets import Static


class WindowChange(Message):
    def can_replace(self, message: Message) -> bool:
        return isinstance(message, WindowChange)


class WindowView(View, layout=VerticalLayout):
    def __init__(
        self,
        widget: RenderableType | Widget,
        *,
        auto_width: bool = False,
        gutter: SpacingDimensions = (0, 0),
        name: str | None = None
    ) -> None:
        layout = VerticalLayout(gutter=gutter, auto_width=auto_width)
        self.widget = widget if isinstance(widget, Widget) else Static(widget)
        layout.add(self.widget)
        super().__init__(name=name, layout=layout)

    async def update(self, widget: Widget | RenderableType) -> None:
        layout = self.layout
        assert isinstance(layout, VerticalLayout)
        layout.clear()
        self.widget = widget if isinstance(widget, Widget) else Static(widget)
        layout.add(self.widget)
        self.layout.require_update()
        self.refresh(layout=True)
        await self.emit(WindowChange(self))

    async def handle_update(self, message: messages.Update) -> None:
        message.prevent_default()
        await self.emit(WindowChange(self))

    async def handle_layout(self, message: messages.Layout) -> None:
        self.log("TRANSLATING layout")
        self.layout.require_update()
        message.stop()
        self.refresh()

    async def watch_virtual_size(self, size: Size) -> None:
        await self.emit(WindowChange(self))

    async def watch_scroll_x(self, value: int) -> None:
        self.layout.require_update()
        self.refresh()

    async def watch_scroll_y(self, value: int) -> None:
        self.layout.require_update()
        self.refresh()

    async def on_resize(self, event: events.Resize) -> None:
        await self.emit(WindowChange(self))



from __future__ import annotations

from typing import Iterable

from ..geometry import Offset, Region, Size, Spacing, SpacingDimensions
from ..layout import Layout, WidgetPlacement
from ..widget import Widget
from .._loop import loop_last


class VerticalLayout(Layout):
    def __init__(
        self,
        *,
        auto_width: bool = False,
        z: int = 0,
        gutter: SpacingDimensions = (0, 0, 0, 0)
    ):
        self.auto_width = auto_width
        self.z = z
        self.gutter = Spacing.unpack(gutter)
        self._widgets: list[Widget] = []
        self._max_widget_width = 0
        super().__init__()

    def add(self, widget: Widget) -> None:
        self._widgets.append(widget)
        self._max_widget_width = max(widget.app.measure(widget), self._max_widget_width)

    def clear(self) -> None:
        del self._widgets[:]
        self._max_widget_width = 0

    def get_widgets(self) -> Iterable[Widget]:
        return self._widgets

    def arrange(self, size: Size, scroll: Offset) -> Iterable[WidgetPlacement]:
        index = 0
        width, _height = size
        gutter = self.gutter
        x, y = self.gutter.top_left
        render_width = (
            max(width, self._max_widget_width)
            if self.auto_width
            else width - gutter.width
        )

        total_width = render_width

        gutter_height = max(gutter.top, gutter.bottom)

        for last, widget in loop_last(self._widgets):
            if (
                not widget.render_cache
                or widget.render_cache.size.width != render_width
            ):
                widget.render_lines_free(render_width)
            assert widget.render_cache is not None
            render_height = widget.render_cache.size.height
            region = Region(x, y, render_width, render_height)
            yield WidgetPlacement(region, widget, (self.z, index))
            y += render_height + (gutter.bottom if last else gutter_height)

        yield WidgetPlacement(Region(0, 0, total_width + gutter.width, y))
